import { isAllOfSchema, isAnyOfSchema, isOneOfSchema, isPropertySchema, isRefSchema, isSchema } from "./type-utils.js";
const SCHEMA_DEFINITION_PREFIX = ['#/definitions/', '#/$defs/'];
function collectRef(def, key, collect) {
    if (isRefSchema(def)) {
        collect(def.$ref, key);
    }
    else if (isSchema(def)) {
        collectRefs(def, key, collect);
    }
}
function collectRefs(def, key, collect) {
    if (isPropertySchema(def)) {
        for (const d of Object.values(def.properties)) {
            collectRef(d, key, collect);
        }
    }
    else if (isAllOfSchema(def)) {
        for (const d of def.allOf) {
            collectRef(d, key, collect);
        }
    }
    else if (isAnyOfSchema(def)) {
        for (const d of def.anyOf) {
            collectRef(d, key, collect);
        }
    }
    else if (isOneOfSchema(def)) {
        for (const d of def.oneOf) {
            collectRef(d, key, collect);
        }
    }
    else if (Array.isArray(def.items)) {
        for (const d of def.items) {
            collectRef(d, key, collect);
        }
    }
    else if (def.items) {
        if (isSchema(def.items) && isRefSchema(def.items)) {
            collect(def.items.$ref, key);
        }
    }
    else if (isRefSchema(def)) {
        collect(def.$ref, key);
    }
}
/**
 * Get all the definitions that are referenced by other definitions.
 */
export function getReferenceMap(definitions) {
    const references = new Map();
    const visited = new Set();
    const addToReferenceMap = (ref, key) => {
        const refKey = ref.replace(SCHEMA_DEFINITION_PREFIX[0], '').replace(SCHEMA_DEFINITION_PREFIX[1], '');
        // Circular reference
        if (key === refKey) {
            return;
        }
        const visitKey = `${key}->${refKey}`;
        if (visited.has(visitKey))
            return;
        visited.add(visitKey);
        const refSet = references.get(key) ?? new Set();
        refSet.add(refKey);
        references.set(key, refSet);
        const refDef = definitions[refKey];
        if (refDef && isSchema(refDef)) {
            collectRefs(refDef, key, addToReferenceMap);
        }
    };
    for (const [key, def] of Object.entries(definitions)) {
        if (isSchema(def)) {
            collectRefs(def, key, addToReferenceMap);
        }
    }
    return references;
}
